/*********************************************************************
 *       This file is part of the Mandala library source code.       *
 *                   Copyright (C) 2002                              *
 *                                                                   *
 *       Authors: eipi - eipiequalsnoone@userOps.sf.net               *
 *                                                                   *
 * This library is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU Lesser General Public        *
 * License as published by the Free Software Foundation; either      *
 * version 2.1 of the License, or (at your option) any later version.*
 *                                                                   *
 * This library is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU *
 * Lesser General Public License for more details.                   *
 *                                                                   *
 * You should have received a copy of the GNU Lesser General Public  *
 * License along with this library; if not, write to                 *
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,   *
 * Boston, MA  02111-1307  USA                                       *
 *********************************************************************/
package org.ow2.novabpm.util;

import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;
import java.util.logging.SimpleFormatter;

/**
 * <p>
 * Custom implementation of the {@link SimpleFormatter} class.
 * <p>
 * 
 * <p>
 * This implementation produce the following pattern:<br>
 * <center><code>L DATE thr#Id Logger Class.Method(): MSG</code></center>
 * 
 * where: <ul>
 * <li>L is a letter code for log level:
 *      <ul>
 *      <li>C for Level.CONFIG;</li>
 *      <li>F for Level.FINE, Level.FINER, Level.FINEST</li>
 *      <li>I for Level.INFO</li>
 *      <li>W for Level.WARNING</li>
 *      <li>S for Level.SEVERE</li>
 *      <li>E for log.entering() or log.exiting() log type</li>
 *      <li>T for log.throwing() log type</li>
 *      </ul>
 * <li>Id is the thread id of the caller</li>
 * </ul>
 * 
 * @author <a href="mailto:eipiequalsnoone@userOps.sf.net">eipi</a>
 * @version $Revision: 1.4 $
 * @since 1.0
 */
public class TraceFormatter extends SimpleFormatter {
  public static final String ALIAS_PROPERTY_KEY = TraceFormatter.class.getName() + ".alias";
  private final Map<String, String> alias;
  private boolean firstTime = true;

  public TraceFormatter() {
    alias = new HashMap<String, String>();
    String aliasConfig = LogManager.getLogManager().getProperty(ALIAS_PROPERTY_KEY);
    if (aliasConfig == null) {
      aliasConfig = System.getProperty(ALIAS_PROPERTY_KEY);
    }
    if (aliasConfig != null) {
      final StringTokenizer configTokenizer = new StringTokenizer(aliasConfig, ",");
      while (configTokenizer.hasMoreTokens()) {
        final String alia = configTokenizer.nextToken();
        final StringTokenizer aliaTokenizer = new StringTokenizer(alia, "~");
        if (aliaTokenizer.countTokens() != 2) {
          throw new IllegalArgumentException("Alias configuration should be of the form 'key1~alia1,key2~alia2,...,keyN~aliaN': " + aliasConfig);
        }
        alias.put(aliaTokenizer.nextToken(), aliaTokenizer.nextToken());
      }
    }
  }

  @Override
  public String format(final LogRecord record) {
    final StringBuffer sb = new StringBuffer(160); // Two 80 colons lines.
    if (firstTime) {
      firstTime = false;
      sb.append(TraceFormatter.class.getName() + " uses the following format:" 
          + Misc.LINE_SEPARATOR
          + "L DATE thr#Id Logger Class.Method(): MSG"
          + Misc.LINE_SEPARATOR
          + "Where L can be: " 
          + Misc.LINE_SEPARATOR
          + "\tC for Level.CONFIG," 
          + Misc.LINE_SEPARATOR
          + "\tF for Level.FINE, Level.FINER and Level.FINEST"
          + Misc.LINE_SEPARATOR
          + "\tI for Level.INFO" 
          + Misc.LINE_SEPARATOR
          + "\tW for Level.WARNING"
          + Misc.LINE_SEPARATOR
          + "\tS for Level.SEVERE" 
          + Misc.LINE_SEPARATOR
          + "\tE for log.entering() or log.exiting() log type"
          + Misc.LINE_SEPARATOR
          + "\tT for log.throwing() log type" 
          + Misc.LINE_SEPARATOR);
      if (alias.isEmpty()) {
        sb.append("Alias can be provided to shorten the output with the property " + ALIAS_PROPERTY_KEY
          + " set to something similar to: string1~alias1,string2~alias2,...");
      } else {
        sb.append("Using alias: " + alias);
      }
          sb.append(Misc.LINE_SEPARATOR + Misc.LINE_SEPARATOR);
    }
    
    final String msg = record.getMessage();
    final boolean entry = msg.startsWith("ENTRY");
    final boolean exit = msg.startsWith("RETURN");
    final boolean throwing = msg.startsWith("THROW");
    if (entry || exit) {
      sb.append('E');
    } else if (throwing){
      sb.append('T');
    } else {
      sb.append(record.getLevel().getName().charAt(0));
    }
    sb.append(' ');
    sb.append(Misc.CONCISE_DATE_FORMATTER.format(record.getMillis()));
    sb.append(" thr#");
    sb.append(record.getThreadID());
    sb.append(' ');
    final String logger = replaceWithAlias(record.getLoggerName());
    sb.append(logger);
    sb.append(' ');
    final String className = replaceWithAlias(record.getSourceClassName());
    sb.append(className);
    sb.append('.');
    sb.append(record.getSourceMethodName());
    sb.append('(');

    if (entry) {
      final Object[] params = record.getParameters();
      if (params != null) {
        final int last = params.length - 1;
        for (int i = 0; i < last; i++) {
          sb.append(params[i]);
          sb.append(", ");
        }
        sb.append(params[last]);
      }
    }
    sb.append(')');
    if (exit || throwing) {
      sb.append(" -> ");
      if (exit) {
        sb.append(record.getParameters()[0]);
      } else {
        sb.append(record.getThrown());
      }
    } else if (!entry) {
      sb.append(": ");
      sb.append(msg);
    }
    sb.append(Misc.LINE_SEPARATOR);

    return new String(sb);
  }

  private String replaceWithAlias(final String src) {
    if (src == null) {
      return src;
    }
    final StringBuilder dst = new StringBuilder(src);
    for (String key : alias.keySet()) {
      final int index = dst.indexOf(key);
      if (index != -1) {
        dst.delete(index, index + key.length());
        dst.insert(index, alias.get(key));
      }
    }
    return new String(dst);
  }
}
